
from sklearn.tree import _tree
from sklearn import tree
import numpy as np
import torch 
import sklearn.metrics as metrics


def getNodesId(ind, Hind):
    ind -= 1                  # index starts from 0
    branchNodes = [ind]
    currentNodes = [ind]
    for _ in range(Hind-1):
        nextNodes = [2*node + j for node in currentNodes for j in [1, 2]]
        branchNodes.extend(nextNodes)
        currentNodes = nextNodes
    leftLeaf = 2*(ind+1) if Hind == 1 else 2*(nextNodes[0]+1)      # read index start from 1. eg, 32 means the 32th node, the first leaf node and index should be 31
    return branchNodes, leftLeaf


## retrieve the parameters abc of the trained tree model
def regTreeWarmStart(model, treeDepth):
    tree_ = model.tree_                  # trained model after fit
    branchNode_inputDepth = 2**(treeDepth) - 1
    # Fitted_treeDepth = model.get_depth()
    # print("Fitted_treeDepth: ", Fitted_treeDepth)
    leafNode_inputDepth = 2**(treeDepth)
    a = [0]*branchNode_inputDepth
    b = [0]*branchNode_inputDepth
    c = [0]*leafNode_inputDepth

    ab0indList = []     # save the ind when a = 0; b = 0
    def warmStartPara(node, ind):
        if tree_.feature[node] != _tree.TREE_UNDEFINED:
            # feature index starts from 0: 0 -> 1st feature 
            featureIdx = tree_.feature[node]
            threshold = tree_.threshold[node]
            a[ind-1] = featureIdx
            # since we use it as warmStart of our NN, we need to change the sign of the threshold
            b[ind-1] = -threshold
            
            node_l = 2 * ind
            node_r = 2 * ind + 1
            warmStartPara(tree_.children_left[node], node_l)
            warmStartPara(tree_.children_right[node], node_r)
        
        else:
            if ind <= branchNode_inputDepth:
                ## node only with one sample. it will be regrarded as a leaf node, but it might be a branch node actually. 
                currDepthForInd = int(np.log2(ind))
                diffDepthInbd = treeDepth - currDepthForInd
                ab0NodeListForEachInd, leftLeaf = getNodesId(ind, diffDepthInbd)
                ab0indList.extend(ab0NodeListForEachInd)
                c[leftLeaf-1-branchNode_inputDepth] = tree_.value[node].squeeze()
            else:
                c[ind-1-branchNode_inputDepth] = (tree_.value[node].squeeze())

    warmStartPara(0, 1)
    return a, b, c, ab0indList


def CARTRegWarmStart(X, Y, treeDepth, device):
    model = tree.DecisionTreeRegressor(max_depth=treeDepth, min_samples_leaf=1, random_state=0)
    if device == torch.device('cuda'):
        X_np, Y_np = X.cpu().numpy(), Y.cpu().numpy()
    else:
        X_np, Y_np = X, Y


    if X_np.shape[0] < 1:
        branchNodeNum = 2**(treeDepth) - 1
        leafNodeNum = 2**(treeDepth)
        p = X_np.shape[1]
        b = [0]*branchNodeNum
        c = [0]*leafNodeNum
        a = np.zeros((branchNodeNum, p), dtype="float32")
        return a, np.asarray(b,dtype="float32"), np.asarray(c, dtype="float32")

    else:
        model = model.fit(X_np, Y_np)
        p = X.shape[1]
        a, b, c, ab0indList = regTreeWarmStart(model,treeDepth)
        a_all = np.eye(p, dtype="float32")[a]                   
        a_all[ab0indList] = 0

        return a_all, np.asarray(b,dtype="float32"), np.asarray(c, dtype="float32")






import math
import torch
from typing import Optional
from torch.optim.lr_scheduler import _LRScheduler


class WarmUpScheduler(_LRScheduler):
    """
    Args:
        optimizer: [torch.optim.Optimizer] only pass if using as astand alone lr_scheduler
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        eta_min: float = 0.0,
        last_epoch=-1,
        max_lr: Optional[float] = 0.1,
        warmup_steps: Optional[int] = 0,
    ):

        if warmup_steps != 0:
            assert warmup_steps >= 0

        self.base_max_lr = max_lr
        self.max_lr = max_lr
        self.step_in_cycle = last_epoch
        self.eta_min = eta_min
        self.warmup_steps = warmup_steps  # warmup

        super(WarmUpScheduler, self).__init__(optimizer, last_epoch)

        self.init_lr()

    def init_lr(self):
        self.base_lrs = []
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = self.eta_min
            self.base_lrs.append(self.eta_min)

    def get_lr(self):
        if self.step_in_cycle == -1:
            return self.base_lrs
        elif self.step_in_cycle < self.warmup_steps:
            return [(self.max_lr - base_lr) * self.step_in_cycle / self.warmup_steps + base_lr
                    for base_lr in self.base_lrs]

        else:
            return [base_lr + (self.max_lr - base_lr) for base_lr in self.base_lrs]

    def step(self, epoch=None):
        self.epoch = epoch
        if self.epoch is None:
            self.epoch = self.last_epoch + 1
            self.step_in_cycle = self.step_in_cycle + 1

        else:
            self.step_in_cycle = self.epoch

        self.max_lr = self.base_max_lr
        self.last_epoch = math.floor(self.epoch)
        for param_group, lr in zip(self.optimizer.param_groups, self.get_lr()):
            param_group['lr'] = lr


class CosineAnealingWarmRestartsWeightDecay(_LRScheduler):
    """
       Helper class for chained scheduler not to used directly. this class is synchronised with
       previous stage i.e.  WarmUpScheduler (max_lr, T_0, T_cur etc) and is responsible for
       CosineAnealingWarmRestarts with weight decay
       """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        T_0: int,
        T_mul: float = 1.,
        eta_min: float = 0.001,
        last_epoch=-1,
        max_lr: Optional[float] = 0.1,
        gamma: Optional[float] = 1.,
    ):

        if T_0 <= 0 or not isinstance(T_0, int):
            raise ValueError("Expected positive integer T_0, but got {}".format(T_0))
        if T_mul < 1 or not isinstance(T_mul, int):
            raise ValueError("Expected integer T_mul >= 1, but got {}".format(T_mul))
        self.T_0 = T_0
        self.T_mul = T_mul
        self.base_max_lr = max_lr
        self.max_lr = max_lr
        self.T_i = T_0  # number of epochs between two warm restarts
        self.cycle = 0
        self.eta_min = eta_min
        self.gamma = gamma
        self.T_cur = last_epoch  # number of epochs since the last restart
        super(CosineAnealingWarmRestartsWeightDecay, self).__init__(optimizer, last_epoch)

        self.init_lr()

    def init_lr(self):
        self.base_lrs = []
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = self.eta_min
            self.base_lrs.append(self.eta_min)

    def get_lr(self):
        return [
            base_lr + (self.max_lr - base_lr) * (1 + math.cos(math.pi * self.T_cur / self.T_i)) / 2
            for base_lr in self.base_lrs
        ]

    def step(self, epoch=None):
        self.epoch = epoch
        if self.epoch is None:
            self.epoch = self.last_epoch + 1
            self.T_cur = self.T_cur + 1
            if self.T_cur >= self.T_i:
                self.cycle += 1
                self.T_cur = self.T_cur - self.T_i
                self.T_i = self.T_i * self.T_mul

        # since warmup steps must be < T_0 and if epoch count > T_0 we just apply cycle count for weight decay
        if self.epoch >= self.T_0:
            if self.T_mul == 1.:
                self.T_cur = self.epoch % self.T_0
                self.cycle = self.epoch // self.T_0
            else:
                n = int(math.log((self.epoch / self.T_0 * (self.T_mul - 1) + 1), self.T_mul))
                self.cycle = n
                self.T_cur = self.epoch - int(self.T_0 * (self.T_mul**n - 1) / (self.T_mul - 1))
                self.T_i = self.T_0 * self.T_mul**(n)

        # base condition that applies original implementation for cosine cycles for details visit:
        # https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CosineAnnealingWarmRestarts.html
        else:
            self.T_i = self.T_0
            self.T_cur = self.epoch

        # this is where weight decay is applied
        self.max_lr = self.base_max_lr * (self.gamma**self.cycle)
        self.last_epoch = math.floor(self.epoch)
        for param_group, lr in zip(self.optimizer.param_groups, self.get_lr()):
            param_group['lr'] = lr


class ChainedScheduler(_LRScheduler):
    """
    Driver class
        Args:
        T_0: First cycle step size, Number of iterations for the first restart.
        T_mul: multiplicative factor Default: -1., A factor increases T_i after a restart
        eta_min: Min learning rate. Default: 0.001.
        max_lr: warmup's max learning rate. Default: 0.1. shared between both schedulers
        warmup_steps: Linear warmup step size. Number of iterations to complete the warmup
        gamma: Decrease rate of max learning rate by cycle. Default: 1.0 i.e. no decay
        last_epoch: The index of last epoch. Default: -1

    Usage:

        ChainedScheduler without initial warmup and weight decay:

            scheduler = ChainedScheduler(
                            optimizer,
                            T_0=20,
                            T_mul=2,
                            eta_min = 1e-5,
                            warmup_steps=0,
                            gamma = 1.0
                        )

        ChainedScheduler with weight decay only:
            scheduler = ChainedScheduler(
                            self,
                            optimizer: torch.optim.Optimizer,
                            T_0: int,
                            T_mul: float = 1.0,
                            eta_min: float = 0.001,
                            last_epoch=-1,
                            max_lr: Optional[float] = 1.0,
                            warmup_steps: int = 0,
                            gamma: Optional[float] = 0.9
                        )

        ChainedScheduler with initial warm up and weight decay:
            scheduler = ChainedScheduler(
                            self,
                            optimizer: torch.optim.Optimizer,
                            T_0: int,
                            T_mul: float = 1.0,
                            eta_min: float = 0.001,
                            last_epoch = -1,
                            max_lr: Optional[float] = 1.0,
                            warmup_steps: int = 10,
                            gamma: Optional[float] = 0.9
                        )
    Example:
        >>> model = AlexNet(num_classes=2)
        >>> optimizer = optim.SGD(model.parameters(), lr=0.1, momentum=0.9, weight_decay=1e-1)
        >>> scheduler = ChainedScheduler(
        >>>                 optimizer,
        >>>                 T_0 = 20,
        >>>                 T_mul = 1,
        >>>                 eta_min = 0.0,
        >>>                 gamma = 0.9,
        >>>                 max_lr = 1.0,
        >>>                 warmup_steps= 5 ,
        >>>             )
        >>> for epoch in range(100):
        >>>     optimizer.step()
        >>>     scheduler.step()

    Proper Usage:
        https://wandb.ai/wandb_fc/tips/reports/How-to-Properly-Use-PyTorch-s-CosineAnnealingWarmRestarts-Scheduler--VmlldzoyMTA3MjM2

    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        T_0: int,
        T_mul: float = 1.0,
        eta_min: float = 0.001,
        last_epoch=-1,
        max_lr: Optional[float] = 1.0,
        warmup_steps: Optional[int] = 5,
        gamma: Optional[float] = 0.95,
    ):

        if T_0 <= 0 or not isinstance(T_0, int):
            raise ValueError("Expected positive integer T_0, but got {}".format(T_0))
        if T_mul < 1 or not isinstance(T_mul, int):
            raise ValueError("Expected integer T_mul >= 1, but got {}".format(T_mul))
        if warmup_steps != 0:
            assert warmup_steps < T_0
            warmup_steps = warmup_steps + 1  # directly refers to epoch account for 0 off set

        self.T_0 = T_0
        self.T_mul = T_mul
        self.base_max_lr = max_lr
        self.max_lr = max_lr
        self.T_i = T_0  # number of epochs between two warm restarts
        self.cycle = 0
        self.eta_min = eta_min
        self.warmup_steps = warmup_steps  # warmup
        self.gamma = gamma
        self.T_cur = last_epoch  # number of epochs since the last restart
        self.last_epoch = last_epoch

        self.cosine_scheduler1 = WarmUpScheduler(
            optimizer,
            eta_min=self.eta_min,
            warmup_steps=self.warmup_steps,
            max_lr=self.max_lr,
        )
        self.cosine_scheduler2 = CosineAnealingWarmRestartsWeightDecay(
            optimizer,
            T_0=self.T_0,
            T_mul=self.T_mul,
            eta_min=self.eta_min,
            max_lr=self.max_lr,
            gamma=self.gamma,
        )

    def get_lr(self):
        if self.warmup_steps != 0:
            if self.epoch < self.warmup_steps:
                return self.cosine_scheduler1.get_lr()
        if self.epoch >= self.warmup_steps:
            return self.cosine_scheduler2.get_lr()

    def step(self, epoch=None):
        self.epoch = epoch
        if self.epoch is None:
            self.epoch = self.last_epoch + 1

        if self.warmup_steps != 0:
            if self.epoch < self.warmup_steps:
                self.cosine_scheduler1.step()
                self.last_epoch = self.epoch

        if self.epoch >= self.warmup_steps:
            self.cosine_scheduler2.step()
            self.last_epoch = self.epoch


import torch
from .treeFunc import readTreePath
from .warmStart import CARTRegWarmStart
from .GET_IterAlp import  multiStartTreeOptbyGRAD_withC
from .subtreePolish import RT


class BaseGETRegressor:
    def __init__(self, treeDepth=4, epochNum=3000, startNum=10, device='cpu'):
        self.treeDepth = treeDepth
        self.epochNum = epochNum
        self.device = torch.device(device)
        self.indices_flags_dict = readTreePath(treeDepth, device)
        self.startNum = startNum
        self.tree = None

    def _prepare_input(self, X):
        if isinstance(X, torch.Tensor):
            return X.clone().detach().to(dtype=torch.float32, device=self.device)
        return torch.tensor(X, dtype=torch.float32, device=self.device)

    def predict(self, X):
        X = self._prepare_input(X)
        n, _ = X.shape
        Tb = 2 ** self.treeDepth - 1
        t = torch.ones(n, device=self.device, dtype=torch.long)

        for _ in range(self.treeDepth):
            decisions = (self.tree['a'][t - 1] * X).sum(dim=1) > self.tree['b'][t - 1]
            t = torch.where(decisions, 2 * t + 1, 2 * t).long()

        Yhat = self.tree['c'][(t - (Tb + 1)).long()]
        return Yhat.detach()


class GETRegressor(BaseGETRegressor):
    def fit(self, X, y):
        X = self._prepare_input(X)
        y = self._prepare_input(y)

        a_init, b_init, c_init = CARTRegWarmStart(X, y, self.treeDepth, self.device)
        cart_warmStart_dict = {"a": a_init, "b": b_init, "c": c_init}
        warmStart = [cart_warmStart_dict]

        _, self.tree = multiStartTreeOptbyGRAD_withC(
            X, y, self.treeDepth, self.indices_flags_dict, self.epochNum, self.device, warmStart, self.startNum
        )
    

class GETSubPolRegressor(BaseGETRegressor):
    def fit(self, X, y):
        X = self._prepare_input(X)
        y = self._prepare_input(y)
        
        _, self.tree, _, _, _ = RT(X, y, self.treeDepth, self.indices_flags_dict, self.epochNum, self.device, self.startNum)


import torch
import math
import h5py
import numpy as np
from typing import Dict, Tuple
import os



# jit version of update_c 
#  tree: Dict[str, torch.Tensor] does not allow any int type value (only tensor tyoe value). 
@torch.jit.script
def update_c(X: torch.Tensor, y: torch.Tensor, treeDepth: int, tree: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
    n, p = X.shape
    Tb = 2 ** treeDepth - 1  # branch node size
    Tleaf = 2 ** treeDepth  # leaf node size
    miny = torch.min(y)
    maxy = torch.max(y)

    # Initialize c and z
    c = torch.full([int(Tleaf)], (miny + maxy) / 2, device=X.device)
    z = torch.ones(n, device=X.device, dtype=torch.long)

    # Calculate the path for each data point in a vectorized manner
    for _ in range(treeDepth):
        # decisions = (tree['a'][z - 1] * X).sum(dim=1) >= tree['b'][z - 1]
        decisions = (tree['a'][z - 1] * X).sum(dim=1) > tree['b'][z - 1]                         # since  ax<=b -> left branch in CART.
        z = torch.where(decisions, 2 * z + 1, 2 * z)
    

    # Adjust indices to match the original function
    z = z - (Tb + 1)
    # Ensure z is of dtype int64
    z = z.to(torch.int64)
    # Calculate the mean of y for each unique element in z
    unique_z, counts = torch.unique(z, return_counts=True)
    # print("unique_z: {}".format(unique_z))
    # print("counts: {}".format(counts))
    sums = torch.zeros_like(c).scatter_add_(0, z, y)
    c[unique_z] = sums[unique_z] / counts.float()
    tree['c'] = c
    return tree




@torch.jit.script
def objv_cost(X: torch.Tensor, y: torch.Tensor, treeDepth: int, tree: Dict[str, torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor]:
    n, p = X.shape
    Tb = 2 ** treeDepth - 1
    t = torch.ones(n, device=X.device, dtype=torch.long)

    for iiii in range(treeDepth):
        # decisions = (tree['a'][t - 1] * X).sum(dim=1) >= tree['b'][t - 1]
        decisions = (tree['a'][t - 1] * X).sum(dim=1) > tree['b'][t - 1]
        t = torch.where(decisions, 2 * t + 1, 2 * t).long()

    Yhat = tree['c'][(t - (Tb + 1)).long()]
    # Manual R2 Score computation
    total_variance = torch.sum((y - torch.mean(y)) ** 2)
    residual_variance = torch.sum((Yhat - y) ** 2)
    r2_score = 1 - (residual_variance / total_variance)

    # return torch.sum((Yhat - y) ** 2)/n, r2_score
    return residual_variance/n, r2_score




## accelerate the getBranchNodes function
def getBranchNodes(ind, prelayer):
    ind -= 1
    branchNodes = [ind]
    currentNodes = [ind]
    for _ in range(prelayer-1):
        nextNodes = [2*node + j for node in currentNodes for j in [1, 2]]
        branchNodes.extend(nextNodes)
        currentNodes = nextNodes
    return branchNodes



## Tree Path Calculation and Saving 
def treePathCalculation(treeDepth, Data_device):
    branchNodeNum = 2 ** (treeDepth) - 1
    leafNodeNum = 2 ** treeDepth  

    ancestorTF_pairs = []              

    for Idx in range(leafNodeNum):
        leafIdx = Idx + branchNodeNum + 1
        log_val = math.floor(math.log2(leafIdx))
        ancestors = [leafIdx >> j for j in range(log_val, -1, -1)]
        ancestors = torch.as_tensor(ancestors, device= Data_device)
        ancestors_shifted = ancestors[1:] + 1
        oddEven = ((-1) ** ancestors_shifted + 1) / 2     # 0 -> I1, 1 -> (1-I1)

        # print("leafIdx: ", leafIdx)
        # print("ancestors: ", ancestors)
        # print("oddEven: ", oddEven)

        ancestorIdxTemp = [(ancestors[ancestorIdx]-1).cpu().numpy().item() for ancestorIdx in range(len(ancestors)-1)]
        oddEvenTemp = [bool(element) for element in ((1- oddEven))]                     # 0/False -> (1-I1), 1/True -> I1

        ancestorTF_zip = list(zip(ancestorIdxTemp, oddEvenTemp))
        ancestorTF_pairs.append(ancestorTF_zip)
    ancestorTF_pairs_np = np.array(ancestorTF_pairs)
    print("ancestorTF_pairs_np: ", ancestorTF_pairs_np.shape)
    print("ancestorTF_pairs_np: ", ancestorTF_pairs_np)

    ## save the ancestorTF_pairs into a HDF5 file
    package_dir = os.path.dirname(__file__)
    ancestor_dir = os.path.join(package_dir, "ancestorTF_File")
    file_path = os.path.join(ancestor_dir, f"ancestorTF_pairs_D{treeDepth}.hdf5")

    with h5py.File(file_path, 'w') as ancestorTF_File:
        # Save the numpy array as a dataset in the HDF5 file
        ancestorTF_File.create_dataset("indicator_pairs", data=ancestorTF_pairs)




## Read Tree Path from HDF5 file
def readTreePath(treeDepth, device):
    ## read the treePath from the HDF5 file
    indices_flags_dict = {}

    package_dir = os.path.dirname(__file__)
    ancestor_dir = os.path.join(package_dir, "ancestorTF_File")

    for treeDepthEach in range(treeDepth):
        treeDepthEach += 1
        file_path = os.path.join(ancestor_dir, f"ancestorTF_pairs_D{treeDepthEach}.hdf5")

        if not os.path.exists(file_path):
            raise FileNotFoundError(f"Missing ancestorTF file: {file_path}. Run treePathCalculation() to generate it.")
        
        with h5py.File(file_path, 'r') as ancestorTF_File:
            ancestorTF_pairs = ancestorTF_File['indicator_pairs'][:]
        
        ancestorTF_pairs_tensor = torch.tensor(ancestorTF_pairs, dtype=torch.long, device=device)

        # Create tensors for indices and flags
        indices_tensor_long = ancestorTF_pairs_tensor[..., 0]
        flags_tensor_long = ancestorTF_pairs_tensor[..., 1]

        key = "D"+str(treeDepthEach)
        indices_flags_dict[key] = {
            'indices_tensor': indices_tensor_long,
            'flags_tensor': flags_tensor_long
        }

    # print(indices_flags_dict.keys())
    return indices_flags_dict






if __name__ == "__main__":

    for treeDepth in [1,2,3,4,5,6,7,8,9,10,11,12]:
        Data_device = torch.device("cpu")
        treePathCalculation(treeDepth, Data_device)


import torch
import copy 
import time
from .warmStart import CARTRegWarmStart
from .treeFunc import update_c, objv_cost, getBranchNodes
from .GET_IterAlp import  multiStartTreeOptbyGRAD_withC


def RT(X, y, treeDepth, indices_flags_dict, epochNum, device, startNum):

    time_GETStart = time.perf_counter()
    aInit, bInit, cInit = CARTRegWarmStart(X, y, treeDepth, device)
    cartWarmStart_dict = {"a": aInit, "b": bInit, "c": cInit}
    warmStart = [cartWarmStart_dict]
    objvMseBase, TreeCombBase = multiStartTreeOptbyGRAD_withC(X, y, treeDepth, indices_flags_dict, epochNum, device, warmStart, startNum)
    time_GETElapsed = time.perf_counter() - time_GETStart
    objv_TreeCombBase, r2_TreeCombBase = objv_cost(X, y, treeDepth, TreeCombBase)
    TreeOpt = copy.deepcopy(TreeCombBase)



    ## Polish the base tree by MH
    n, p = X.shape
    Tb = 2 ** treeDepth - 1                            # branch node size                     
    Tleaf = 2 ** treeDepth                             # leaf node size        
    a = torch.zeros((Tb, p), device= device, dtype=torch.float32)
    b = torch.zeros((Tb), device= device, dtype=torch.float32)

    updateIndList = []

    a, b, TreeOpt, objv_TreeOpt  = RT_inner(X, y, X, y, 1, treeDepth, indices_flags_dict, a, b, 1, epochNum, device, startNum, TreeOpt, objv_TreeCombBase, updateIndList)


    treeDepth_tensor = torch.tensor(treeDepth, dtype=torch.float32, device=device)
    TreeSubPolComb = {"a": a, "b": b, "c": torch.zeros(Tleaf, dtype=torch.float32, device=device), "D": treeDepth_tensor}
    TreeSubPolComb = update_c(X, y, treeDepth_tensor, TreeSubPolComb)
    objv_TreeSubPolComb, _ = objv_cost(X, y, treeDepth, TreeSubPolComb)

    if objv_TreeOpt < objv_TreeSubPolComb:
        return objv_TreeOpt, TreeOpt, objv_TreeCombBase, TreeCombBase, time_GETElapsed
    else:
        return objv_TreeSubPolComb, TreeSubPolComb, objv_TreeCombBase, TreeCombBase, time_GETElapsed




def RT_inner(X, y, X_all, y_all, dc, Dmax, indices_flags_dict, a_s, b_s, ind, epochNum, device, startNum, TreeOpt, objvOpt, updateIndList):

    TreeBasePolNode = copy.deepcopy(TreeOpt)

    H_ind = Dmax-dc+1

    BranchNodesList = getBranchNodes(ind, H_ind)

    a_branchNodesWS = TreeOpt["a"][BranchNodesList, :]
    b_branchNodesWS = TreeOpt["b"][BranchNodesList]              
    Tleaf = 2 ** H_ind
    tree_branchNodeWS =  {"a": a_branchNodesWS, "b": b_branchNodesWS, "c": torch.zeros(Tleaf, device=device, dtype=torch.float32)}
    tree_branchNodeWS = update_c(X_all, y_all, H_ind, tree_branchNodeWS)
    tree_branchNodeWS = {"a": tree_branchNodeWS["a"], "b": -tree_branchNodeWS["b"], "c": tree_branchNodeWS["c"]}   # negative b here 

    a_init, b_init, c_init = CARTRegWarmStart(X, y, H_ind, device)
    cart_warmStart_dict = {"a": a_init, "b": b_init, "c": c_init}
    warmStart = [cart_warmStart_dict, tree_branchNodeWS]
    objv_iter, tree = multiStartTreeOptbyGRAD_withC(X, y, H_ind, indices_flags_dict, epochNum, device, warmStart, startNum)


    a_s[ind-1, :] = tree["a"][0, :]
    b_s[ind-1] = tree["b"][0]


    TreeBasePolNode["a"][BranchNodesList, :] = tree["a"]    
    TreeBasePolNode["b"][BranchNodesList] = tree["b"]
    TreeBasePolNode = update_c(X_all, y_all, Dmax, TreeBasePolNode)
    objv_TreeBasePolNode, r2_TreeBasePolNode = objv_cost(X_all, y_all, Dmax, TreeBasePolNode)


    if objv_TreeBasePolNode < objvOpt:
        TreeOpt = copy.deepcopy(TreeBasePolNode)
        objvOpt = objv_TreeBasePolNode
        updateIndList.append(ind)
    else:
        pass
    

    stopDepthForMH = 4
    if Dmax > stopDepthForMH: 
        if dc == stopDepthForMH: 
            a_s[BranchNodesList, :] = tree["a"]
            b_s[BranchNodesList] = tree["b"]
            return a_s, b_s, TreeOpt, objvOpt
    else:
        # reach the depth 
        if dc == Dmax:
            return a_s, b_s, TreeOpt, objvOpt

    

    # node index for next layer 
    node_l = 2 * ind
    node_r = 2 * ind + 1
    yes = torch.matmul(X, a_s[ind-1, :]) <= b_s[ind-1]
    if y[yes].shape[0] > 1 and torch.unique(y[yes]).numel() > 1:
        a_s, b_s, TreeOpt, objvOpt = RT_inner(X[yes, :], y[yes], X_all, y_all, dc + 1, Dmax, indices_flags_dict, a_s, b_s, node_l, epochNum, device, startNum, TreeOpt, objvOpt, updateIndList)
    else:
        # print("y[yes].shape[0] <= 1")
        pass
    if y[~yes].shape[0] > 1 and torch.unique(y[~yes]).numel() > 1:
        a_s, b_s, TreeOpt, objvOpt = RT_inner(X[~yes, :], y[~yes], X_all, y_all, dc + 1, Dmax, indices_flags_dict, a_s, b_s, node_r, epochNum, device, startNum, TreeOpt, objvOpt, updateIndList)
    else:
        # print("y[~yes].shape[0] <= 1")
        pass


    return a_s, b_s, TreeOpt, objvOpt





from setuptools import setup, find_packages

with open("README.md", "r") as f:
    description = f.read()

setup(
    name="get-oblique",
    version="0.1.0",
    include_package_data=True,
    packages=find_packages(),
    package_data={
        "GET": ["ancestorTF_File/*.hdf5"],
    },
    install_requires=[
        "h5py==3.13.0",
        "numpy==1.26.4",
        "pandas==2.2.3",
        "scikit_learn==1.5.0",
        "torch>=2.0.0"
    ],
    long_description=description,
    long_description_content_type="text/markdown"
)


import torch 
import copy 
import math

from .treeFunc import objv_cost, update_c

from .modifiedScheduler import ChainedScheduler


import os 
#  enforce deterministic behavior in PyTorch
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False
torch.use_deterministic_algorithms(True)
os.environ["CUBLAS_WORKSPACE_CONFIG"]=":4096:8"


class scaledSigmoid(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, scale):
        ctx.save_for_backward(input)
        ctx.scale = scale
        sigmoid = torch.sigmoid(scale * input)
        ctx.sigmoid = sigmoid
        return sigmoid

    @staticmethod
    def backward(ctx, grad_output):
        input = ctx.saved_tensors
        scale = ctx.scale
        sigmoid = ctx.sigmoid
        # sigmoid = torch.sigmoid(scale * input)
        gradient = grad_output * (scale * sigmoid * (1 - sigmoid))
        return gradient, None


class branchNodeNet(torch.nn.Module):
    def __init__(self, treeDepth: int, p: int, scale: float) -> None:
        super().__init__()
        self.depth = treeDepth
        self.featNum = p
        # self.treesize = 2 ** (self.depth + 1) - 1
        self.branchNodeNum = 2 ** (self.depth) - 1
        self.scale = scale

        self.linear1 = torch.nn.Linear(self.featNum, self.branchNodeNum)
        self.scaledSigmoid = scaledSigmoid.apply 
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear1(x)
        x = self.scaledSigmoid(-x, self.scale)
        return x



@torch.jit.script
def objectiveFuncwithC(branchOutput: torch.Tensor, Y_train: torch.Tensor, c_leafLable: torch.Tensor,  treeDepth: int, indices_tensor: torch.Tensor, flags_tensor: torch.Tensor) -> torch.Tensor: 

    indicators_complement = 1 - branchOutput
    Y_train_diff = (Y_train.view(-1, 1) - c_leafLable.view(1, -1)).pow_(2)
    indicators_stack = torch.stack((indicators_complement, branchOutput))
    selected_indicators = indicators_stack[flags_tensor, :, indices_tensor]
    indicator_pairs = selected_indicators.prod(dim=1).transpose(0, 1)
    objv = (indicator_pairs * Y_train_diff).sum()
    N_num = Y_train.shape[0]
    objv = objv / N_num 
    return objv




def initialize_parameters(choice, net, warmStarts, Y_train, device):

    a, b, c = None, None, None
    if choice < len(warmStarts) and warmStarts[choice] is not None:
        # print("warm start")
        # print("choice is {}".format(choice))
        ws = warmStarts[choice]
        a, b, c = ws["a"], ws["b"], ws["c"]

    else:
        # print("random init")
        a = torch.rand(net.linear1.weight.shape,dtype=torch.float32, device=device) * (2.0)+(-1.0)
        b = torch.rand(net.linear1.bias.shape, dtype=torch.float32, device=device) * (2.0)+(-1.0)
        c = torch.rand([2 ** net.depth], dtype=torch.float32, device=device) * (torch.max(Y_train) - torch.min(Y_train)) + torch.min(Y_train)

    return a, b, c


class callbackFuncs:
    def CalMSEonEpochEnd(self, X_train, Y_train, treeDepth, net, c_leafLable):
        with torch.no_grad():    
            # Clone the weights and biases to avoid modifying the original tensors
            a_grad = net.linear1.weight.clone().detach()  # Ensuring no gradients
            b_grad = net.linear1.bias.clone().detach() * (-1.0)
            c_grad = c_leafLable.clone().detach()  # Ensuring no gradients
            treeEpoch = {"a": a_grad, "b": b_grad, "c": c_grad}
        treeEpoch = update_c(X_train, Y_train, treeDepth, treeEpoch)
        objvMSE_Epoch, r2_Epoch = objv_cost(X_train, Y_train, treeDepth, treeEpoch)
        return objvMSE_Epoch, r2_Epoch, treeEpoch
        



def treeOptbyGRADwithC(treeDepth, indices_flags_dict, epochNum, X_train, Y_train, device, warmStarts, scaleFactor, lrscheList,  idxStart, callback):

    ## hyperparameters
    learningRate, T0, warmupsteps, gamma = lrscheList[0], lrscheList[1], lrscheList[2], lrscheList[3]

    ##  net
    p = X_train.shape[1]
    scale = torch.tensor([scaleFactor], device=device)
    net = branchNodeNet(treeDepth, p, scale).to(device, non_blocking=True)

    ### initialize weight and bias of net
    a, b, c = initialize_parameters(idxStart, net, warmStarts, Y_train, device)
    customWeigt = torch.as_tensor(a, dtype=torch.float32, device=device)
    customBias = torch.as_tensor(b, dtype=torch.float32, device=device)
    net.linear1.weight = torch.nn.Parameter(customWeigt)
    net.linear1.bias = torch.nn.Parameter(customBias)
    ## optimized parameters
    c_leafLable = torch.as_tensor(c, dtype=torch.float32, device=device).requires_grad_()

    ## Optimizer 
    # optimizer = torch.optim.Adam(list(net.parameters())+[c_leafLable], lr=learningRate)
    optimizer = torch.optim.AdamW(list(net.parameters())+[c_leafLable], lr=learningRate)

    # ############################################
    # ########### if to disable the scheduler, must comment the following code; otherwise, the lr will be affected ############ 
    # restartNum = 1
    # # Sn = 2**restartNum-1
    # Sn = restartNum
    # warmupsteps = 10
    # T0= math.ceil((epochNum - warmupsteps)/Sn)
    # # print("T0 is {} and warmupsteps is {}".format(T0, warmupsteps))
    # epochNumNew = T0*Sn+warmupsteps
    # # print("epochNumNew is {}".format(epochNumNew))
    # gamma = (1/restartNum)**(1/(restartNum-1))   if restartNum != 1 else 1
    # # print("gamma is {}".format(gamma))
    scheduler = ChainedScheduler(optimizer, T_0=T0, T_mul=1, eta_min=1e-6, max_lr=learningRate, warmup_steps=warmupsteps, gamma=gamma)

    # load the indices_tensor and flags_tensor from the indices_flags_dict
    indices_tensor = indices_flags_dict["D"+str(treeDepth)]["indices_tensor"]
    flags_tensor = indices_flags_dict["D"+str(treeDepth)]["flags_tensor"]


    objvMSE_EpochBest = float('inf')
    r2_EpochBest = float('-inf')
    tree_EpochBest = None

    for epoch in range(epochNum):

        optimizer.zero_grad(set_to_none=True)   

        # Forward pass with a batch
        branchOutput = net(X_train)
        objv = objectiveFuncwithC(branchOutput, Y_train, c_leafLable, treeDepth, indices_tensor, flags_tensor)

        ## check the real mse loss of the current tree
        objvMSE_Epoch, r2_Epoch, treeEpoch = callback.CalMSEonEpochEnd(X_train, Y_train, treeDepth, net, c_leafLable)
        if objvMSE_Epoch < objvMSE_EpochBest:
            objvMSE_EpochBest = objvMSE_Epoch
            r2_EpochBest = r2_Epoch
            tree_EpochBest = copy.deepcopy(treeEpoch)

        # Backward pass and optimize
        objv.backward()

        optimizer.step()
        scheduler.step()

    return objvMSE_EpochBest, r2_EpochBest, tree_EpochBest






def multiStartTreeOptbyGRAD_withC(X_train, Y_train, treeDepth, indices_flags_dict, epochNum, device, warmStarts, startNum):

    objvmin = 1e10
    treeOpt = None


    ############################################
    ########### if to disable the scheduler, must comment the following code; otherwise, the lr will be affected ############ 
    restartNum = 1
    # Sn = 2**restartNum-1
    Sn = restartNum
    warmupsteps = 10
    T0= math.ceil((epochNum - warmupsteps)/Sn)
    # print("T0 is {} and warmupsteps is {}".format(T0, warmupsteps))
    epochNumNew = T0*Sn+warmupsteps
    # print("epochNumNew is {}".format(epochNumNew))
    gamma = (1/restartNum)**(1/(restartNum-1))   if restartNum != 1 else 1
    # print("gamma is {}".format(gamma))
    lrscheList = [0.01, T0, warmupsteps, gamma]
    ############################################
    
    cartWarmStart = warmStarts[0]
    treeCART = {"a": torch.tensor(cartWarmStart["a"], device=device), "b": torch.tensor(-cartWarmStart["b"], device=device), "c": torch.tensor(cartWarmStart["c"], device=device)}
    objvMSECART, r2CART = objv_cost(X_train, Y_train, treeDepth, treeCART)


    # callback function
    callback = callbackFuncs()


    if len(warmStarts) > startNum:
        startNum = len(warmStarts)
    # print("startNum is {}".format(startNum))
    for idxStart in range(startNum):
        scaleStart = torch.rand(1, device=device) * 20 + 5
        scaleEnd = torch.rand(1, device=device) * 40 + 80
        scaleList = [scaleStart.item(), scaleEnd.item()]


        ObjvAlp = 1e10
        bestTreeAlp = None
        warmStarts_cur = [warmStarts[idxStart]] if idxStart < len(warmStarts) else  [None]
        idxCurr = 0
        

        for scaleFactor in scaleList:

            objvCurr, r2Curr, treeCurrent = treeOptbyGRADwithC(treeDepth, indices_flags_dict, epochNum, X_train, Y_train, device, warmStarts_cur, scaleFactor, lrscheList, idxCurr, callback)
            if objvCurr < ObjvAlp:
                ObjvAlp = objvCurr
                bestTreeAlp = copy.deepcopy(treeCurrent)
            warmStarts_cur.append( {"a": treeCurrent["a"], "b": treeCurrent["b"]* (-1.0), "c": treeCurrent["c"]})
            idxCurr += 1

        if ObjvAlp < objvMSECART:
            # TreeAfterGrad = copy.deepcopy(bestTreeAlp)
            TreeAfterGrad = bestTreeAlp
        else:
            # TreeAfterGrad = copy.deepcopy(treeCART)
            TreeAfterGrad = treeCART
            ObjvAlp = objvMSECART
        
        if ObjvAlp < objvmin:
            objvmin = ObjvAlp
            treeOpt = copy.deepcopy(TreeAfterGrad)

    return objvmin, treeOpt
























